Miscellaneous utilities on ordinary Haskell syntax used by the arrow
translator.

> module Utils(
>	FreeVars(freeVars), DefinedVars(definedVars),
>	failureFree, irrPat, paren, parenInfixArg,
>	tuple, tupleP,
>	postorderPat
> ) where

> import Data.Set
> import Language.Haskell.Syntax

The set of free variables in some construct.

> class FreeVars a where
>	freeVars :: a -> Set HsName

> instance FreeVars a => FreeVars [a] where
>	freeVars = unionManySets . map freeVars

> instance FreeVars HsPat where
>	freeVars (HsPVar n) = unitSet n
>	freeVars (HsPLit _) = emptySet
>	freeVars (HsPNeg p) = freeVars p
>	freeVars (HsPInfixApp p1 _ p2) = freeVars p1 `union` freeVars p2
>	freeVars (HsPApp _ ps) = freeVars ps
>	freeVars (HsPTuple ps) = freeVars ps
>	freeVars (HsPList ps) = freeVars ps
>	freeVars (HsPParen p) = freeVars p
>	freeVars (HsPRec _ pfs) = freeVars pfs
>	freeVars (HsPAsPat n p) = freeVars p `addToSet` n
>	freeVars (HsPWildCard) = emptySet
>	freeVars (HsPIrrPat p) = freeVars p

> instance FreeVars HsPatField where
>	freeVars (HsPFieldPat _ p) = freeVars p

> instance FreeVars HsFieldUpdate where
>	freeVars (HsFieldUpdate _ e) = freeVars e

> instance FreeVars HsExp where
>	freeVars (HsVar n) = freeVars n
>	freeVars (HsCon _) = emptySet
>	freeVars (HsLit _) = emptySet
>	freeVars (HsInfixApp e1 op e2) =
>		freeVars e1 `union` freeVars op `union` freeVars e2
>	freeVars (HsApp f e) = freeVars f `union` freeVars e
>	freeVars (HsNegApp e) = freeVars e
>	freeVars (HsLambda _ ps e) = freeVars e `minusSet` freeVars ps
>	freeVars (HsLet decls e) =
>		(freeVars decls `union` freeVars e) `minusSet`
>			definedVars decls
>	freeVars (HsIf e1 e2 e3) =
>		freeVars e1 `union` freeVars e2 `union` freeVars e3
>	freeVars (HsCase e as) = freeVars e `union` freeVars as
>	freeVars (HsDo ss) = freeVarsStmts ss
>	freeVars (HsTuple es) = freeVars es
>	freeVars (HsList es) = freeVars es
>	freeVars (HsParen e) = freeVars e
>	freeVars (HsLeftSection e op) = freeVars e `union` freeVars op
>	freeVars (HsRightSection op e) = freeVars op `union` freeVars e
>	freeVars (HsRecConstr _ us) = freeVars us
>	freeVars (HsRecUpdate e us) = freeVars e `union` freeVars us
>	freeVars (HsEnumFrom e) = freeVars e
>	freeVars (HsEnumFromTo e1 e2) = freeVars e1 `union` freeVars e2
>	freeVars (HsEnumFromThen e1 e2) = freeVars e1 `union` freeVars e2
>	freeVars (HsEnumFromThenTo e1 e2 e3) =
>		freeVars e1 `union` freeVars e2 `union` freeVars e3
>	freeVars (HsListComp e ss) =
>		freeVars e `union` freeVarsStmts ss
>	freeVars (HsExpTypeSig _ e _) = freeVars e
>	freeVars (HsAsPat _ _) = error "freeVars (x @ p)"
>	freeVars (HsWildCard) = error "freeVars _"
>	freeVars (HsIrrPat _) = error "freeVars ~p"

> instance FreeVars HsQOp where
>	freeVars (HsQVarOp n) = freeVars n
>	freeVars (HsQConOp _) = emptySet

> instance FreeVars HsQName where
>	freeVars (UnQual v) = unitSet v
>	freeVars _ = emptySet

> instance FreeVars HsAlt where
>	freeVars (HsAlt _ p gas decls) =
>		(freeVars gas `union` freeVars decls) `minusSet`
>		(freeVars p `union` definedVars decls)

> instance FreeVars HsGuardedAlts where
>	freeVars (HsUnGuardedAlt e) = freeVars e
>	freeVars (HsGuardedAlts alts) = freeVars alts

> instance FreeVars HsGuardedAlt where
>	freeVars (HsGuardedAlt _ e1 e2) = freeVars e1 `union` freeVars e2

> instance FreeVars HsDecl where
>	freeVars (HsFunBind ms) = freeVars ms
>	freeVars (HsPatBind _ p rhs decls) =
>		(freeVars rhs `union` freeVars decls) `minusSet`
>		(freeVars p `union` definedVars decls)
>	freeVars _ = emptySet

> instance FreeVars HsMatch where
>	freeVars (HsMatch _ n ps rhs decls) =
>		(freeVars rhs `union` freeVars decls) `minusSet`
>		((freeVars ps `addToSet` n) `union` definedVars decls)

> instance FreeVars HsRhs where
>	freeVars (HsUnGuardedRhs e) = freeVars e
>	freeVars (HsGuardedRhss grs) = freeVars grs

> instance FreeVars HsGuardedRhs where
>	freeVars (HsGuardedRhs _ e1 e2) = freeVars e1 `union` freeVars e2

> freeVarsStmts :: [HsStmt] -> Set HsName
> freeVarsStmts = foldr addStmt emptySet
>	where	addStmt (HsGenerator _ p e) s =
>			freeVars e `union` (s `minusSet` freeVars p)
>		addStmt (HsQualifier e) _s = freeVars e
>		addStmt (HsLetStmt decls) s =
>			(freeVars decls `union` s) `minusSet` definedVars decls

The set of variables defined by a construct.

> class DefinedVars a where
>	definedVars :: a -> Set HsName

> instance DefinedVars a => DefinedVars [a] where
>	definedVars = unionManySets . map definedVars

> instance DefinedVars HsDecl where
>	definedVars (HsFunBind (HsMatch _ n _ _ _:_)) = unitSet n
>	definedVars (HsPatBind _ p _ _) = freeVars p
>	definedVars _ = emptySet

Is the pattern failure-free?
(This is incomplete at the moment, because patterns made with unique
constructors should be failure-free, but we have no way of detecting them.)

> failureFree :: HsPat -> Bool
> failureFree (HsPVar _) = True
> failureFree (HsPApp n ps) = n == unit_con_name && null ps
> failureFree (HsPTuple ps) = all failureFree ps
> failureFree (HsPParen p) = failureFree p
> failureFree (HsPAsPat _ p) = failureFree p
> failureFree (HsPWildCard) = True
> failureFree (HsPIrrPat _) = True
> failureFree _ = False

Irrefutable version of a pattern

> irrPat :: HsPat -> HsPat
> irrPat p@(HsPVar _) = p
> irrPat (HsPParen p) = HsPParen (irrPat p)
> irrPat (HsPAsPat n p) = HsPAsPat n (irrPat p)
> irrPat p@(HsPWildCard) = p
> irrPat p@(HsPIrrPat _) = p
> irrPat p = HsPIrrPat p

Make an expression into an aexp, by adding parentheses if required.

> paren :: HsExp -> HsExp
> paren e = if isAexp e then e else HsParen e
>	where	isAexp (HsVar _) = True
>		isAexp (HsCon _) = True
>		isAexp (HsLit _) = True
>		isAexp (HsParen _) = True
>		isAexp (HsTuple _) = True
>		isAexp (HsList _) = True
>		isAexp (HsEnumFrom _) = True
>		isAexp (HsEnumFromTo _ _) = True
>		isAexp (HsEnumFromThen _ _) = True
>		isAexp (HsEnumFromThenTo _ _ _) = True
>		isAexp (HsListComp _ _) = True
>		isAexp (HsLeftSection _ _) = True
>		isAexp (HsRightSection _ _) = True
>		isAexp (HsRecConstr _ _) = True
>		isAexp (HsRecUpdate _ _) = True
>		isAexp _ = False

Make an expression into an fexp, by adding parentheses if required.

> parenInfixArg :: HsExp -> HsExp
> parenInfixArg e@(HsApp _ _) = e
> parenInfixArg e = paren e

Tuples

> tuple :: [HsExp] -> HsExp
> tuple [] = unit_con
> tuple [e] = e
> tuple es = HsTuple es

> tupleP :: [HsPat] -> HsPat
> tupleP [] = HsPApp unit_con_name []
> tupleP [e] = e
> tupleP es = HsPTuple es

Monadic post-order traversal of a pattern.
Unlike a pre-order version, this is guaranteed to terminate on a
finite pattern if the argument function does.

> postorderPat :: Monad m => (HsPat -> m HsPat) -> HsPat -> m HsPat
> postorderPat post = traverse
>	where	traverse p = do
>			p' <- traverseSubpats p
>			post p'
>		traverseSubpats p@(HsPVar _) =
>			return p
>		traverseSubpats p@(HsPLit _) =
>			return p
>		traverseSubpats (HsPNeg p) = do
>			p <- traverse p
>			return (HsPNeg p)
>		traverseSubpats (HsPInfixApp p1 e p2) = do
>			p1 <- traverse p1
>			p2 <- traverse p2
>			return (HsPInfixApp p1 e p2)
>		traverseSubpats (HsPApp n ps) = do
>			ps <- mapM traverse ps
>			return (HsPApp n ps)
>		traverseSubpats (HsPTuple ps) = do
>			ps <- mapM traverse ps
>			return (HsPTuple ps)
>		traverseSubpats (HsPList ps) = do
>			ps <- mapM traverse ps
>			return (HsPList ps)
>		traverseSubpats (HsPParen p) = do
>			p <- traverse p
>			return (HsPParen p)
>		traverseSubpats (HsPRec n pfs) = do
>			let	postorderField (HsPFieldPat n p) = do
>					p <- traverse p
>					return (HsPFieldPat n p)
>			pfs <- mapM postorderField pfs
>			return (HsPRec n pfs)
>		traverseSubpats (HsPAsPat n p) = do
>			p <- traverse p
>			return (HsPAsPat n p)
>		traverseSubpats p@HsPWildCard =
>			return p
>		traverseSubpats (HsPIrrPat p) = do
>			p <- traverse p
>			return (HsPIrrPat p)
