-- A few simple Pan examples

module Simple where

import Monad
import Pan

-- We can create "UIs" that encapsulate interface and effect, and combine
-- them in a modular way.  Note that Filter c = Image c -> Image c, so
-- these UIs yield values that are image-transforming functions.

rotateUI, rippleUI,swirlUI  :: UI (Filter a)
cropUI, swirlRippleCropUI :: UI FilterC  -- crop is monomorphic

rotateUI = do angle <- fullAngleUI
              return $ rotate angle

rippleUI = do n <- islider "# ripples" (1,15) 3
              d <- slider  "depth" (-0.5,1.5) 0.3
              return $ rippleRadius n d

cropUI   = do r <- slider "crop radius" (0,200) 40
              return $ crop (disk r)

swirlUI  = do swirliness <- slider "swirliness" (-2,2) 0.5
              return $ swirl (100/swirliness)

-- Composite effect.  Ripple-then-crop or crop-then-ripple.  Swirl result.
swirlRippleCropUI = do rippleFirst <- tick "Ripple before crop" False
                       swirlF      <- swirlUI
                       rippleF     <- rippleUI
                       cropF       <- cropUI
                       return $
                         swirlF . ifE rippleFirst (cropF.rippleF) (rippleF.cropF)


-- Note the common pattern in rotateUI, cropUI, and swirlUI: a value is
-- extracted from a UI and used to construct a new value, which is then
-- returned.  Haskell is very good at being able to abstract patterns, and
-- in fact already has a suitable monadic combinator.  Using liftM, we get
-- the following simpler definitions:

rotateUI'   = liftM rotate $ fullAngleUI

rippleUI' = liftM2 rippleRadius
              (islider "# ripples" (1,15) 3)
              (slider  "depth" (-0.5,1.5) 0.3)

swirlUI'  = liftM (swirl . (100/))
                  (slider "swirliness" (-2,2) 0.5)

powerScaleUI :: UI (Filter a)
powerScaleUI = liftM powScale $ slider "power" (-2,2) 0.5
 where
   powScale  z im = im . powScaleP (1/z)
   powScaleP z = polarXf (\ (r,a) -> ((r ** z) / (50 ** (z-1)), a))

spinUI :: UI (Image c -> ImageB c)
spinUI = liftM (\ r im t -> rotate (-r*t) im) (slider "rate" (-5,5) 2)

-- We could use liftM4 and a complicated function to reformulate
-- swirlRippleCropUI, but it seems more clear as is.  On the other hand,
-- we might factor out the combination of rippleFirst, rippleF, cropF into
-- a single UI, and then compose with swirlUI with the result as in
-- tileTurn below.

-- From now on, we'll use monadic style where it's convenient.


wedgeAnnulus :: FloatE -> IntE -> Region
wedgeAnnulus inner arms = uscale 50 (annulus inner) `intersectR` radReg arms

wedgeAnnulusUI :: UI Region
wedgeAnnulusUI = liftM2 wedgeAnnulus
                       (slider "inner radius" (0,1) 0.5)
                       (islider "arms" (0,50) 3)

wedgeAnnulusSwirlUI = swirlUI `ap` wedgeAnnulusUI

---- Tiling.

tileUI :: UI (Filter a)
tileUI = liftM tile sizeUI

-- Turn the image being tiled.  Just compose functions from rotateUI and tileUI
tileRotateUI :: UI (Filter a)
tileRotateUI = tileUI `comp` rotateUI

fadeUI :: UI FilterC
fadeUI = liftM (fade . const) $ slider "opacity" (0,1) 0.5

fadeOverUI, tileRotateFadeUI :: UI ImageCombiner

fadeOverUI = do fadeF <- fadeUI
                return $ \ imA imB _ -> fadeF imA `over` imB

tileRotateFadeUI = do comb       <- fadeOverUI
                      tileRotate <- tileRotateUI
                      return $ \ imA imB t -> comb (tileRotate imA) imB t

-- Don't use imageSource, because the source and result may have very
-- different sizes.
imageToTileUI :: UI SizedImageC
imageToTileUI = browseImageSize "image to tile" "flower.bmp"

tileSizeXfGUI :: (VectorE -> Filter ColorE) -> UI FilterC
              -> UI SizedImageC -> UI ImageC
tileSizeXfGUI tiler xfUI simUI = 
  do sep      <- liftM (2**) $ slider "separation power" (-1,1) 0
     rotateF    <- xfUI
     (w,h,im) <- simUI
     return $ tiler (uscaleP sep (w,h)) $ rotateF im

tileSizeXfUI :: UI FilterC -> UI SizedImageC -> UI ImageC
tileSizeXfUI = tileSizeXfGUI tile

tileSizeUI, tileRotateSizeUI :: UI SizedImageC -> UI ImageC
tileSizeUI                   = tileSizeXfUI (return id)
tileRotateSizeUI             = tileSizeXfUI rotateUI

-- This one is self-sizing
tileUI' :: UI ImageC
tileUI' = tileSizeUI imageToTileUI

-- Alternating rings of true/false
altRings :: Region
altRings p = evenE (floorE (distO p))

altRingsUI :: UI Region
altRingsUI = liftM (flip uscale altRings)
                   (slider "ring size" (0,100) 20)


-- Advanced example.  ToDo: move out of this "Simple" module.
{-
 Histograms the brightness of an input image. Not useful on its own,
 but demonstrates how you can use the Action monad to do imperative programming
 & then use 'performAction' to make it appear functional to the outside
 world.
-}
histoUI :: UI ImageC
histoUI = do
  (w,h,im) <- imageSourceSize "input image"
  let 
      arrIn = pixelize (f2i w,f2i h) im
	-- construct a histogram of the lightness of the input image.
      hist  = 
  	histogram (arr2Map (\ (r,g,b,a) -> f2i (i2f (r+g+b) / 3)) arrIn)

	-- visualise the histogram in a 255x255 array, which we then
	-- 'reconstruct' and display.
      output = 
        Array2 255 255 imf 
	where
         imf (x,y) = let v = readArr hist x  in
  	             ifE (y <* v) black invisible
  return (reconstruct output)
  

histogram :: Array2 ByteFracE -> Exp (Arr Int)
histogram arrIn = 
	performAction (
	    newArr 255 0 >>=* \ arr ->
	    loop (arr2Height arrIn) ( \ j ->
  	     loop (arr2Width arrIn) ( \ i ->
	      let
	       r   = indexArray2 arrIn (i,j)
	       old = readArr arr r
	      in
	      writeArr arr r (old + 1))) >>*
	   returnE arr)

{-
 Translate and scale a brightness range of the input image.
-}
contrastUI = do
  il <- slider "source start" (0,1) 0
  ih <- slider "source end" (0,1) 1
  ol <- slider "target start" (0,1) 0
  oh <- slider "target end" (0,1) 1
  return (\ im -> (\ p -> let c = im p in brightXf (il,ih) (ol,oh) c))

brightXf :: (FractionE, FractionE) -> (FractionE, FractionE) -> ColorE -> ColorE
brightXf (il,ih) (ol,oh) (b1,g1,r1,a1) = (h i, h i, h i, a1)
 where
   i = (b1 + g1 + r1) `div` 3
   h bf = 
        let v = fromBF bf in
   	ifE (bf >=* toBF il &&* bf <=* toBF ih)
   	    (toBF (ol + (oh-ol)*((v-il) / (ih-il))))
	    bf

-- Random examples, better put these in some Test module instead.

hsvDemo :: UI ImageC
hsvDemo = do
--  dh  <- slider "hue" (0,360) 0
--  ds  <- slider "saturation" (0,1) 0.1
  dv  <- slider "brightness" (-1,1) 0.1
  im <- imageSource "input image" 
  let inRange mn mx v = max mn (min v mx)
  return (\ (x,y) -> let (h,s,v,a) = rgbToHSV (im (x,y)) in hsvToRGB (h,s, inRange 0 1 (dv+v), a))

randDemo :: UI ImageC
randDemo = do
    f <- islider "factor" (1,256) 1
    sz <- islider "square size" (1,10) 5
    im@(Array2 w h imf) <- imageSourceD "input image" 
    let
      mkArr' size elemAction = 
          performAction (
	   unsafeNewArr size >>=* \ arr ->
	   loop size (\ i -> elemAction i >>=* \ v -> writeArr arr i v) >>*
	   returnE arr)

    let arrE =
	 mkArr' (w*h) (\ _ -> rand >>=* \ v -> returnE (v `mod` f))
--	 mkArr' (w*h) (\ _ -> rand >>=* \ v -> returnE (fromBF (v `mod` f)))
	within mn mx x = max mn (min mx x)
--	arr = Array2 w h (\ (x,y) -> fadeC (1 - readArr arrE (x*y)) (imf (x,y)))
	arr = Array2 w h (\ (x,y) ->
			  let
			    dx = readArr arrE ((x `div` sz))
			    dy = readArr arrE (((x+1) `div` sz))
			  in
			  imf (within 0 w (x-dx),within 0 h (y+dy)))
    return (reconstruct arr)
      
brightShift = do 
  im <- imageSource "input image"
  am <- slider "displacement" (1,200) 64
  return (\ p@(x,y) -> let v = lightness (im p) in im (x,y-am*v))

displacementUI = do
   im <- imageSource "input image"
   am <- slider "displacement" (-1,1) 0
   return (\ t -> im . translateP (0.1 * cos (am * t), 0.1 * sin (am*t)))

----------------------------------------

-- This command builds stand-alone and plug-in versions of all of the
-- examples in this module.  Use "dispFilter" to build just the
-- stand-alone, or "plugFilter" for just the PhotoShop plug-in.
-- "makeFilter" builds both.  The trivial argument is to prevent CAFs.

makeAll = makeAllFor allBackends

makeAllFor bends = do
  compileFilterFor  bends "rotate"            $ atMouse rotateUI
  compileFilterBFor bends "spin"              $ spinUI
  compileFilterFor  bends "ripple"            $ atMouse rippleUI
  compileFilterFor  bends "crop"              $ atMouse cropUI
  compileFilterFor  bends "swirl"             $ atMouse swirlUI
  compileFilterFor  bends "swirl ripple crop" $ atMouse swirlRippleCropUI
  compileFilterFor  bends "power scale"       $ atMouse powerScaleUI
  compileImageFor   bends "tile rotate"       $ tileRotateUI `ap` imageSource "input image"
  compileCombinerFor bends "fade over"        fadeOverUI
  -- Too much, and not very interesting.  Better if the tiled & faded were
  -- a PNG with transparency.
  -- makeCombiner "tile turn fade" tileTurnFadeUI
