
module Cse(cse,rename) where

import List(partition,nub,(\\),intersect,sort)
import Char(ord)
import IOExts
import Types
import F32vec4 (F32vec4)
import DExp
import FiniteMap
import Maybe
import CseUtils( collect,topsort,gensym
               , oin,notin,twice,mergecount
	       , merge,myintersect,twicein
	       , insert,newNode, Node
	       , ptrToInt, intToNode
	       )
import Exp (mkE)                         -- for pretty printing
import TCheck (tcheck)
import qualified Cost

---------------------------------------------------------------------------------------------------------
-- node-labelled graphs

type Graph a = FiniteMap Node ([Node],a)

children :: Graph a -> Node -> [Node]
children g n = case lookupFM g n of
                 Just (ns,_) -> ns
                 Nothing     -> []

edges :: Graph a -> [(Node,Node)]
edges g = [(a,b) | (a,(bs,_)) <- fmToList g, b<-bs]

addedge :: Graph a -> (Node,Node) -> Graph a
addedge g (a,b) = 
     addToFM_C (\ (x,a) (y,b) -> (y++x,a) )
               g a
	       ([b],error "undefined in Share.addedge")

labelsonly :: Graph a -> Graph a
labelsonly = mapFM (\ _ (_,b) -> ([],b))

converse :: Graph a -> Graph a
converse f =  foldl addedge (labelsonly f) (map swap (edges f))
              where swap (a,b) = (b,a)

mapNodes :: ([Node] -> a -> b) -> Graph a -> Graph b
mapNodes f = mapFM (\ _ (ns,a) -> (ns,f  ns a))

mapGraph :: (Node -> [Node] -> a -> b) -> Graph a -> Graph b
mapGraph f = mapFM (\ n (ns,a) -> (ns,f n ns a))



scan :: (Graph lab -> (lab,b) -> Node -> [a] -> a) -> Graph (lab,b) -> Graph ((lab,b),a)
scan f g = g'
           where g' = mapGraph (\ n ns a -> (a,f g'' a n (scans ns))) g
                 g'' = mapNodes (\ ns (a,b) -> a) g
                 scans = map (attr. label . lookupWithDefaultFM g' (error "undefd"))
                 label (n,l) = l
                 attr (l,a) = a 

sscan :: (lab -> [a] -> a) -> Graph lab -> Graph (lab,a)
sscan f g = g'
            where g' = mapGraph (\ n ns a -> (a,f a (scans ns))) g
                  scans = map (attr. label . lookupWithDefaultFM g' (error "undefd"))
                  label (n,l) = l
                  attr (l,a) = a 

-- tag the nodes with their max height (only makes sense for acyclic graphs)
labelHeight :: Graph a -> Graph (a,Int)
labelHeight g = g'
                where g'          = mapNodes f g
                      f ns a      = (a, 1+ foldr max (-1) (depths ns))
                      depths      = map (depth.label.lookupWithDefaultFM g' ([],(error "undefined in Share.labeldepth",-1)))
                      depth (a,d) = d
                      label (n,l) = l


labelTop :: Graph a -> Graph (a,Int)
labelTop g = mapFM (\ n (ns,a) -> (ns,(a, topnumber n)) ) g
             where topnumber = lookupWithDefaultFM top (error "undefined in Share.labelTop.topnumber") 
                   top = listToFM (zip (map node (topsortgraph g)) [0..])
                   node (n,l) = n

topsortgraph :: Graph a -> [(Node,a)]
topsortgraph g = reverse (topsort [ (a,b,as) | (a,(as,b)) <- fmToList g])


-- Turn a dag into a minimal dag: so we've got the same result as if we
-- were hash-consing.

compact :: Ord a 
	=> Graph a 
	-> Node
	-> (Graph a, FiniteMap (a,[Node]) Node, FiniteMap Node Node)
	-> (Node, Graph a, FiniteMap (a,[Node]) Node, FiniteMap Node Node)
compact h n (g,f,v)= case lookupFM v n of
                      Just k -> (k,g,f,v)
                      Nothing -> (nm,ng,nf,nv) 
                     where (ns,l)    = case lookupFM h n of
                                         Just (ns,l) -> (ns,l)
                                         Nothing -> error "undefined in compact"
                           (ms,mg,mf,mv)  = foldr visit ([],g,f,v) ns
                           (nm,ng,nf,nv)  = case lookupFM f (l,ms) of
                                              Just k -> (k,mg,mf,addToFM mv n k) 
                                              Nothing -> let k = newNode () in
                                                          (k,addToFM mg k (ms,l),addToFM mf (l,ms) k,addToFM mv n k)
                           visit n (ms,g,f,v) = let (m,ng,nf,nv) = compact h n (g,f,v) in (m:ms,ng,nf,nv)

compress :: Ord a => Graph a  -> Node -> (Graph a,Node)
compress g n = (ng,k)
               where (k,ng,_,_ ) = compact g n (emptyFM,emptyFM,emptyFM) 

sizeGraph :: Graph a -> Int
sizeGraph g = length (fmToList g)

traceGraph :: Show a => Graph a -> Graph a
traceGraph g = trace (show (fmToList g) ++ "\n") g 
-------------------------------------------------------------------------------------------------------
-- labels for representing Exps as labelled graphs

data Label = Op Op
           | LitInt' Int
           | LitDouble' Double
           | LitBool' Bool
           | ImpData' String
           -- | Let' Id Exp Label
           | If'
           | Loop' Id  Node Node
           | MkArr' Id  Node Node
	   | NewArr' Type Node Node
           | Lambda' [(Id,Type)]
           | Var' Id Type            deriving (Eq,Ord,Show)

-- which labels are nullary?
atomic LitInt'{}       = True
atomic LitDouble'{}    = True
atomic LitBool'{}      = True
--atomic LitF32vec4'{} = True -- False  -- Encourage sharing
atomic ImpData'{}      = True
atomic Var'{}          = True
atomic Lambda'{}       = True  -- why?
atomic _               = False

-- given a label and a list of expressions, build corresponding expression
fromHead :: Label -> [Exp] -> Exp

fromHead (Op op) ts      = fromHnf (op,ts) 
fromHead (LitInt' n) _   = LitInt n
fromHead (LitDouble' n) _ = LitDouble n
fromHead (LitBool' n) _  = LitBool n
fromHead (ImpData' x) _  = ImportData x
-- fromHead (Let' x t l) ts = Let x t (fromHead l ts)
fromHead (Lambda' xs) [e] = foldr (\ (x,ty) acc -> Lambda x ty acc) e xs
fromHead (Var' x t) _     = Var x t      
fromHead If' [c,t,e]      = If c t e
fromHead (Loop' x _ _) [t,e]  = Loop x t e
fromHead (MkArr' x _ _) [n,e]  = MkArr n x e
fromHead (NewArr' ty _ _) [n]  = NewArr n ty
fromHead l es = error ("missing case in fromHead" ++ show l ++ show es )

-- turn an expression into a labelled graph: this relies on the unsafe primitive "ptrToInt"
-- we use a depth-first search on the expression, stopping when a subexpression has been 
-- encountered previously. It is assumed that there are no "Let"s in the original expression.

graph :: Exp -> Node -> Graph Label
graph e n = gr e n emptyFM

gr :: Exp -> Node -> Graph Label -> Graph Label
gr  (LitInt n) pv f       = addToFM f pv ([],LitInt' n)
gr  (LitDouble n) pv f    = addToFM f pv ([],LitDouble' n)
gr  (LitBool b) pv f      = addToFM f pv ([],LitBool' b)
gr  (ImportData d) pv f   = addToFM f pv ([],ImpData' d)
gr  (Var x t) pv f        = addToFM f pv ([],Var' x t)
gr  e@(Lambda x t b) pv f = 
	let
	 grabLambdas (Lambda x t b) = let (as,b') = grabLambdas b in ((x,t) : as, b')
	 grabLambdas b = ([],b)

	 (vs,b') = grabLambdas b
	in
	grs f pv e (Lambda' ((x,t):vs)) [b']
gr  e@(If c t s) pv f     = grs f pv e If' [c,t,s]
gr  e@(Loop x c b) pv f   = let zn = intToNode 0 in grs f pv e (Loop' x zn zn) [c,b]
gr  e@(MkArr c x b) pv f  = let zn = intToNode 0 in grs f pv e (MkArr' x zn zn) [c,b]
gr  e@(NewArr sz ty) pv f = let zn = intToNode 0 in grs f pv e (NewArr' ty zn zn) [sz]
gr  (Let _ _ _) pv f      = error "cannot take graph of let"
gr  e pv f                = grs f pv e (Op l) ts
                          where (l,ts) = toHnf e

grs f pv e l ts = case (lookupFM f pv) of
                Nothing -> foldr ( \ (x,y) acc -> gr x y acc) (addToFM f pv (vs1,l)) vs
                Just _  -> f
  where
    vs = map (\ x -> (x, ptrToInt x)) ts
    vs1 = map snd vs
    

-- turn a labelled graph into an expression. Takes graph and "entry node" as arguments


expr :: Graph Label -> Node -> Exp
expr f n = case (lookupFM f n) of
             Just (ns,l) -> fromHead l (map (expr f) ns)
             Nothing     -> error "undefined link in Share.expr"

------------------------------------------------------------------------------------------------------------------
tracePhases = True
checkPhases = False
showFinal   = False

phase :: String -> (Exp -> Exp) -> (Exp -> Exp)
phase name f e =
  -- Evaluate e using size, then show phase name, then
  -- compute and type-check phase's result.
  (if tracePhases then phaseTrace else id) $
  (if checkPhases then phaseCheck else id) $
  result
 where
   phaseTrace = s `seq` trace ("Phase \"" ++ name ++
                               "\" generated "
                               ++ show s ++ " nodes.\n")
   phaseCheck = tcheck ("Phase " ++ name)
   result = f e
   s = siz result


done | showFinal = \ e -> siz e `seq`
                          trace ("Final CSEd exp:\n" ++ show (mkE e) ++ "\n\n") e
     | otherwise = id

cse :: Exp -> Exp
cse = done
    . phase "cleanup"     simplify
    . phase "share"       share
    . phase "simplify"    id

compressphase :: (Graph Label, Node) -> (Graph Label,Node)
compressphase (g,n) = if tracePhases
                      then let (g',n') = compress g n
                               s1 = sizeGraph g
                               s2 = sizeGraph g'
                           in s1 `seq` s2 `seq` 
                              trace ("Phase \"compress\": size " ++ show s1 ++ " compressed to " ++ show s2 ++ ".\n") (g',n')
                      else compress g n

-----------------------------------------------------------------------------------------------------------------
-- share: expose sharing with lets

share :: Exp -> Exp
share e = expand qv $       -- place the lets
          -- traceGraph $
          analyse $         -- find potential placements for the lets
          -- traceGraph $
          mark $            -- mark candidates for let-abstraction and hoisting
          freelabel $       -- label each node with the free variables below it
          annotLoops $      -- annotate each loop/mkarray with nodes of boundvar and zero
          dag
          where (dag,qv) = compressphase (graph e pv,pv)
                pv = ptrToInt e



----------------------------------------------------------------------------------------------------------------
-- annotLoops: annotate loop/mkarray labels with boundvar and zero nodes

annotLoops :: Graph Label -> Graph Label
annotLoops g = mapNodes f g
               where f _ (Loop' x _ _)    = Loop' x (nvar x) zero 
                     f _ (MkArr' x _ _)   = MkArr' x (nvar x) zero
                     f _ l          = l
                     zero = nodeOfZero g
                     nvar = nodeOfVar g

nodeOfZero g = if null xs then intToNode 0 else head xs
               where xs = [  n |  (n,(ns,LitInt' 0)) <- fmToList g]
nodeOfVar g x = if null xs then intToNode (-(sum (map ord x))) else head xs
                where xs = [ n | (n,(ns,Var' y t)) <- fmToList g, y==x]

-----------------------------------------------------------------------------------------------------------------
-- freelabel: label each node with the free variables below it
 
-- JULY 8: remove anything from this list that is bound by a lambda
globalPanVars = []
--globalPanVars = ["output", "t", "height","output","panX","panY","t","width","zoom"]

freelabel :: Graph Label -> Graph (Label,[Id])
freelabel = sscan frees
            where frees l xs = case l of
                                 Var' x t     -> if elem x  globalPanVars then [] else [x]
                                 Loop' x _ _  -> foldr merge [] xs \\ [x]
                                 MkArr' x _ _ -> foldr merge [] xs \\ [x]
                                 Lambda' vs   -> foldr merge [] xs \\ map fst vs
                                 _            -> foldr merge [] xs

------------------------------------------------------------------------------------------------------------------
{- mark: mark candidates for let-abstraction and hoisting
   the result is a labelled graph, where beside the label itself, each node is annotated with
   a triple (repeated subexpression to be abstracted?, 
             expression to be hoisted?,
             free variables)                  
-}

mark :: Graph (Label,[Id]) -> Graph (Label,(Bool,Bool,[Id]))
mark g = mapGraph (\n ns (l,ys) -> (l, let res = m n in res)) g
         where m n = case lookupFM g' n of
                       Just (ns,(l,xs)) -> let --yss = map ids ns
                                               nonAction = not (isActionTy (typeOf (expr g'' n)))
--                                               nonAction = typeOf (expr g'' n) /= Action
                                           in
                                                 (length ns > 1 && not (atomic l) && nonAction,
                                                  -- more than one parent and not trivial and not an action
                                                  not (atomic l) && 
                                                  nonAction && 
                                                  --(or (map (super xs) yss) || not (null (loop l))),
                                                  (or (map (parentTest xs) ns) || not (null (loop l))),
                                                  -- not trivial, not an action, and at least one parent has a
                                                  -- super set of free variables
                                                  xs)
                       _                -> error "undefined in Share.mark.m"
               g' = converse g
               g'' = mapNodes (\ns (l,xs) -> l) g -- unfortunate for efficiency: pass types in as node labels?
               ids n = case lookupFM g' n of
                         Just (ns,(l,xs)) -> (l,xs)
                         _                -> error "undefined in Share.mark.ids"

               -- For nontrivial, non-actions, should hoist if parent has superset of free vars, or is an action.
               parentTest xs p = super xs (ids p) || hasActionTy g'' p

loop :: Label -> [Id]
loop (Loop'  x _ _) = [x]
loop (MkArr' x _ _) = [x]
loop (Lambda' xs)   = map fst xs
loop _              = []

super :: [Id] -> (Label,[Id]) -> Bool
super xs (l,ys) = sort xs /= sort ys' && and (map (flip elem ys') xs)
                  where ys' = lx ++ (ys \\ lx)
                        lx  = loop l

hasActionTy :: Graph Label -> Node -> Bool
hasActionTy g n = isActionTy (typeOf (expr g n))

------------------------------------------------------------------------------------------------------------------

analyse :: Graph (Label,(Bool,Bool,[Id])) -> Graph (Label,[Node],[(Node,[Constraint])],[(Hoist,[Constraint])])
analyse g = (mapNodes (\ns ((l,b),o) -> (l,locals o,elocals o,hlocals o)) . scan (step g)) g

------------------------------------------------------------------------------------------------------------------
-- CONSTRAINTS

assumeMkArrTotal = True

data Constraint = Less Node Node | Leq Node Node | Unprovable deriving (Eq,Ord,Show)

{- expToConstraint reduces a boolean expression and its negation
   to a set of constraints. We require that

                  expToConstraint e => e1 

   and

                  not (expToConstraint e) => e2 -}


implies :: Constraint -> Constraint -> Bool
implies (Less n1 n2) (Leq m1 m2) = n1 == m1 && n2 == m2
implies _ Unprovable = False
implies c1 c2 = c1==c2 

expToConstraint :: (Graph Label, Node) -> ([Constraint],[Constraint])
expToConstraint (g,n) = case lookupFM g n of
                          Just ([n1,n2],Op LessThanOp) -> 
                                 if typeOf (expr g n1) == Int
                                 then  ([Less n1 n2],[Leq n2 n1])
                                 else  ([],[])
                          Just ([n1,n2],Op OrOp) ->
                                 let (cs1,ds1) = expToConstraint (g,n1)
                                     (cs2,ds2) = expToConstraint (g,n2)
                                 in (cs1 `myintersect` cs2, 
                                     ds1 `merge` ds2)
                          Just ([n1,n2],Op AndOp) ->
                                 let (cs1,ds1) = expToConstraint (g,n1)
                                     (cs2,ds2) = expToConstraint (g,n2)
                                 in (cs1 `merge` cs2, 
                                     ds1 `myintersect` ds2)
                          Just ([n1],Op NotOp) ->
                                 let (cs,ds) = expToConstraint (g,n1)
                                 in (ds,cs)
                          Just _ -> ([],[])
                          Nothing -> error "expToConstraint"


-- building constraints

{- ifthen takes a condition and two sets of constraints
   (one for the then-branch, one for the else-branch),
   and produces a new list of constraints, simplifying
   each of the branches using the condition -}

type GExp = (Graph Label, Node)

ifthen :: GExp -> [Constraint] -> [Constraint] -> [Constraint] -> [Constraint]
ifthen e csIf csThen csElse = csIf `merge` simplifyC ds1 csThen `merge` simplifyC ds2 csElse
                              where (ds1,ds2) = expToConstraint e 

{- rangeloop takes a condition and a set of constraints,
   and simplifies the constraints using the condition -}

rangeloop :: Id -> Node -> Node -> Node -> [Constraint] -> [Constraint] -> [Constraint]
rangeloop x nx z n csSize csBody = csSize `merge` simplifyC [Less nx n, Leq z nx] csBody

rangearray :: Node -> Node -> Node -> [Constraint] -> [Constraint] -> [Constraint]
rangearray nx z n csSize csBody = if assumeMkArrTotal
                                    then []
                                    else csSize `merge` simplifyC [Less nx n, Leq z nx] csBody

{- simplifyC takes two sets of constraints, and tries to discharge
   constraints in the second set, assuming the first -}

simplifyC :: [Constraint] -> [Constraint] -> [Constraint]
simplifyC cs ds = filter (not . impl) ds
                  where impl d = or (map (flip implies d) cs)

{- satisfied tests whether a set of constraints is trivially true -}

satisfied :: [Constraint] -> Bool
satisfied = null

sat :: [Constraint] -> (a,[Constraint]) -> Bool
sat cs (n,ds) = satisfied (simplifyC cs ds) 

newconstraints :: Graph (Label,(Bool,Bool,[Id])) -> Label -> Node -> [Constraint]
newconstraints g l n = case l of
                         Op (ReadArrOp _) -> case lookupFM g n of
                                              Just ([n1,n2],(Op (ReadArrOp _),_)) -> case sizer g 0 n1 of
                                                                                      Just (z,s) -> [Less n2 s,Leq z n2]
                                                                                      Nothing -> [Unprovable] 
                                              _ -> error "missing case in Cse.newconstraints" 
                         Op Deref8Op -> [Unprovable]
                         Op Deref32Op -> [Unprovable]
                         _ -> []

sizer :: Graph (Label,(Bool,Bool,[Id])) -> Int -> Node -> Maybe (Node,Node)
sizer g y n = case lookupFM g n of
                 Just ([n1,n2],(MkArr' x nx z,_)) -> sizer' g y [] n 
		 Just ([n1,n2],(NewArr' ty _ _,_)) -> sizer' g y [] n 
                 Just ([n1,n2],(Op (ReadArrOp _),_)) -> sizer g (y+1) n1 
		 Just ([], (Var' x _,_)) -> sizer' g y [] n
		 Just ls -> Nothing -- error ("missing case in Cse.sizer " ++ show ls)
                 _ -> error "missing case in Cse.sizer"

sizer' :: Graph (Label,(Bool,Bool,[Id])) -> Int -> [Id] -> Node -> Maybe (Node,Node)
sizer' g y bs n = case lookupFM g n of
                    Just ([n1,n2],(MkArr' x nx z,_)) -> if y==0  
                                                        then case lookupFM g n1 of
                                                               Just (_,(_,(_,_,fvs))) -> if null (intersect bs fvs)
                                                                                         then Just (z,n1) else Nothing
                                                        else sizer' g (y-1) (x:bs) n2
                    Just ([n1,n2],(NewArr' ty nx z,_)) -> if y==0  
                                                        then case lookupFM g n1 of
                                                               Just (_,(_,(_,_,fvs))) -> if null (intersect bs fvs)
                                                                                         then Just (z,n1) else Nothing
                                                        else sizer' g (y-1) bs n2
		    Just ([], (Var' ty nx,_)) -> Nothing -- 
		    x -> error ("missing case in Cse.sizer1 " ++ show x)
                    _ -> error "missing case in Cse.sizer1"
             
globalConstraints :: Graph Label -> [Constraint]
globalConstraints g = map f globalPanVars
                      where f v = Leq zero (nodeOfVar g v)
                            zero = nodeOfZero g

---------------------------------------------------------------------------------------------------------
-- hoistables

type Hoist = (Node,[Id],[(Id,Node)])  -- exp, free vars, bound vars with upperbound

bump :: Id -> Node -> (Hoist,[Constraint]) -> (Hoist,[Constraint])
bump x m vs@((n,xs,ys),cs) = if elem x xs then ((n,xs \\ [x], (x,m):ys),cs) else vs

bumpLam :: [(Id,Type)] -> (Hoist,[Constraint]) -> Maybe (Hoist,[Constraint])
bumpLam ls vs@((n,xs,ys),cs) = if any (\ x -> elem (fst x) xs) ls then Nothing else Just vs

hoistable :: [Id] -> Hoist -> Bool
hoistable ys (n,xs,_) = and (map (flip elem ys) xs)
---------------------------------------------------------------------------------------------------------

data Out = Out { busy        :: [Node],
                 repeated    :: [Node],
                 cost        :: Int,
                 constraints :: [Constraint],
                 expensive   :: [(Node,[Constraint])],
                 hoist       :: [(Hoist,[Constraint])],
                 locals      :: [Node],
                 hlocals     :: [(Hoist,[Constraint])],
                 elocals     :: [(Node,[Constraint])] } deriving Show

step :: Graph (Label,(Bool,Bool,[Id])) -> Graph Label -> (Label,(Bool,Bool,[Id])) -> Node -> [Out] -> Out
step g' g (l,(x0,x1,xs)) n childrenOut
  = parentOut
    where parentOut = Out {locals      = (oin (busy parentOut) .    -- it's "busy" in this node
                                          twicein .                 -- it occurs in at least two children
                                          map repeated ) childrenOut,
                           busy        = busy' `myintersect` (repeated parentOut),
                           repeated    = if x0 then insert n repeated' else repeated',
                           cost        = 1 + sum (map cost childrenOut),
                           constraints = case l of
                                           If'          -> let [c,_,_] = children g n 
                                                               [csIf,csThen,csElse] = map constraints childrenOut
                                                           in ifthen (g,c) csIf csThen csElse
                                           MkArr' x nx z -> let [sz,_] = children g n
                                                                [csSize,csBody] = map constraints childrenOut
                                                            in rangearray nx z sz csSize csBody 
                                           NewArr' x nx z -> let [sz,_] = children g n
                                                                 [csSize,csBody] = map constraints childrenOut
                                                            in rangearray nx z sz csSize csBody 
					    -- 
					    -- 
					    -- hack: really want to restrict this to applications of Perform
                                            -- that return Array values.
					    --
                                           Op PerformOp | assumeMkArrTotal -> []
                                           Loop' x  nx z -> let [up,_] = children g n
                                                                [csUp,csBody] = map constraints childrenOut
                                                            in rangeloop x nx z up csUp csBody 
                                           _        -> newconstraints g' l n `merge` 
                                                       foldr merge [] (map constraints childrenOut),
                           expensive   = if x0 && cost parentOut > 5    -- repeated, nontrivial and quite large
                                         then insert (n,constraints parentOut) expensive'
                                         else expensive',
                           elocals     = (twicein . map expensive) childrenOut,
                           hoist       = let hoist'' = case l of
                                                         MkArr' x _ _ -> (map (bump x (head (children g n))) hoist')
                                                         Loop' x _ _  -> (map (bump x (head (children g n))) hoist')
							 Lambda' xs   -> mapMaybe (bumpLam xs) hoist'
							 _            -> hoist'
                                         in (if x1 && cost parentOut > 1 -- hoistable, and not entirely trivial
                                            then insert ((n,xs,[]),constraints parentOut) hoist'' else hoist''),
                           hlocals     = case l of
                                           MkArr' _ _ _ -> filter (hoistable xs.fst) hoist'
                                           Loop' _  _ _ -> filter (hoistable xs.fst) hoist'
                                           Lambda' _    -> filter (hoistable xs.fst) hoist'
                                           _        -> [] }
          busy'      = case l of
                         If'          -> let [c,t,e] = map busy childrenOut in
                                         insert n ( c `merge` (t `myintersect` e))
                         Loop' _ _ _  -> let [c,b] = map busy childrenOut in
                                         insert n c    -- don't let through anything at all from the body: refine this
                         MkArr' _ _ _ -> let [s,b] = map busy childrenOut in
                                         insert n s
                         Lambda' _    -> [n]
                         _            -> insert n ( (foldr merge [] . map busy) childrenOut)
          repeated'  = (foldr merge [] . map repeated) childrenOut
          expensive' = case l of
                         Loop' _  _ _ -> let [c,b] = map expensive childrenOut in c
                         MkArr' x nx z -> let [s,b] = map expensive childrenOut in s
                         Lambda' _    -> []
                         _            -> (foldr merge [] . map expensive) childrenOut
          hoist'     = (foldr merge [] . map hoist) childrenOut

expand :: Node
       -> Graph (Label,[Node],[(Node,[Constraint])],[(Hoist,[Constraint])])
       -> Exp
expand n g  = letexpr g (globalConstraints g') emptyFM n
              where g' = mapNodes (\ _ (l,_,_,_) -> l) g


-- letexpr takes a graph, a set of constraints (reflecting assumptions about the context), a
-- finite map that gives the nodes to be abstracted at each location, and
-- a finite map that records abstractions in the environment. It returns the letexpr corresponding
-- to the graph.

letexpr :: Graph (Label,[Node],[(Node,[Constraint])],[(Hoist,[Constraint])]) -> 
           [Constraint] -> 
           FiniteMap Node Exp -> 
           Node -> 
           Exp
letexpr f cs varmap n = case lookupFM varmap n of
                           Just e     -> e 
                           Nothing    -> letexpr' f cs varmap n
                              
letexpr' f cs varmap n 
    = case lookupFM f n of
        Just(_,(l,ns,ts,hs))-> let ls             = filter (not . flip elemFM varmap) (ls2 `merge` ns)
                                   ls2            = (map fst . filter (sat cs)) ts
                                   hs2            = (map mkbins . 
                                                     filter (\(n,_,_) -> not (elemFM n varmap)) .
                                                                         map fst . filter (sat cs)) hs
                                   mkbins (n,_,bs)= let (e,x,t) = ((foldr mkbin (letexpr' f cs varmap' n) bs) ,
                                                                       gensym (),
                                                                       typeOf e)
                                                        i = foldl mkidx (Var x t) bs
                                                    in (n,x,e,i,freevars e)
                                   mkbin  (x,m) e  = MkArr (letexpr' f cs varmap' m) x e 
                                   mkidx i (x,_)  = ReadArr i (Var x Int)
                                   xts            = reverse (topsort (map (\ (n,x,e,i,f) -> (x,e,f)) ls'))
                                   ls'            = map prep ls ++ hs2
                                   prep n         = let x = gensym ()
                                                        e = letexpr' f cs varmap' n
                                                    in (n,x, e, Var x (typeOf e), freevars e)
                                   varmap'        = addListToFM varmap vs
                                   vs             = map (\ (n,x,e,i,f) -> (n,i)) ls'
                                   lt (x,t) b     = Let x t b 
                                in (foldr lt (rec varmap') xts) 
        Nothing -> rec varmap
      where rec vm = case (lookupFM f n) of
                       Just (ns,(l,_,_,_)) -> case l of
                                                  If' -> let [c,t,e] = ns
                                                             g = mapNodes (\ ns (l,_,_,_) -> l) f   -- pass this in?
                                                             (tcs,ecs) = expToConstraint (g,c) in
                                                         (fromHead l [letexpr f cs vm c,
                                                                      letexpr f (tcs `merge` cs) vm t,
                                                                      letexpr f (ecs `merge` cs) vm e])
                                                  MkArr' x nx z -> let [s,b] = ns
                                                                       ss = [Less nx s, Leq z nx]
                                                                   in fromHead l [letexpr f cs vm s, letexpr f (ss `merge` cs) vm b]
                                                  Loop' x nx z -> let [s,b] = ns
                                                                      ss = [Less nx s, Leq z nx]
                                                                   in fromHead l [letexpr f cs vm s, letexpr f (ss `merge` cs) vm b]
                                                  _ -> fromHead l (map (letexpr f cs vm) ns)
                       Nothing     -> error "undefined link in Share.letexpr.rec"



simplify :: Exp -> Exp

simplify e = fst (simple emptyFM e)
simple _ e@(LitInt n)       = (e,[])
simple _ e@(LitDouble f)    = (e,[])
simple _ e@(LitBool b)      = (e,[])
simple _ e@(ImportData _)   = (e,[])
simple f e@(Var n t)        = ((case lookupFM f n of
                                 Nothing -> e 
                                 Just m  -> m), [(n,1)])
simple f (Lambda x ty e)    = (Lambda x ty e', filter ((/=x).fst)xs)
                              where (e',xs) = simple f e
simple f (Let x e1 e2)      = let (e1',xs1) = simple f e1 
                                  (e2',xs2) = simple g e2
                                  keepLet   = twice x xs2 && Cost.worthAbstracting f x e1'
                                  (g,e')    = if keepLet then (f,Let x e1' e2') 
                                                         else (addToFM f x e1',e2')
                              in (e',xs1 `mergecount` xs2)
simple f (If e1 e2 e3)      = (If e1' e2' e3', (xs1 `mergecount` xs2) `mergecount` xs3)
                              where (e1',xs1) = simple f e1
                                    (e2',xs2) = simple f e2
                                    (e3',xs3) = simple f e3
simple _ e@Noop             = (e,[])
simple f (Loop x e1 e2)     = (Loop x e1' e2', filter ((/=x).fst) (xs1 `mergecount` xs2))
                              where (e1',xs1) = simple f e1
                                    (e2',xs2) = simple f e2

-- push Lets outwards, simplifies ReadArr/MkArr matching below.
simple f (ReadArr (Let x rhs body) i) = simple f (Let x rhs (ReadArr body i))
simple f e@(ReadArr arr i) = 
  let (arr',xs1) = simple f arr
      (i'  ,xs2) = simple f i

      e' = mkReadArr arr'

      mkReadArr arrExpr
        = --siz arr' `seq` trace ("simple ReadArr: arr' is " ++ show (mkE arr') ++ "\n") $
          case arrExpr of
            MkArr size ivar body -> --trace "simple ReadArr: hit\n" $
                                      mkLet ivar i' body
	     -- don't let any Lets intervene ReadArr/MkArr cancellation.
	    Let x rhs body       ->   mkLet x rhs (mkReadArr body)
            _                    -> --trace "simple ReadArr: miss\n" $
                                       ReadArr arr' i'
  in (e',xs1 `mergecount` xs2)
 where
   mkLet x (Var x' _) body | x == x' = body
   mkLet x rhs body                  = Let x rhs body

simple f (MkArr e1 x e2)    = (MkArr e1' x e2', filter ((/=x).fst) (xs1 `mergecount` xs2))
                              where (e1',xs1) = simple f e1
                                    (e2',xs2) = simple f e2
simple f e                  = (fromHnf (head, es'), xs)
                              where (head,args) = toHnf e
                                    (es',xss) = unzip (map (simple f) args)
                                    xs = foldr mergecount [] xss

rename :: Id -> Id -> Exp -> Exp
rename old new = renamer (addToFM emptyFM old new)

renamer _ e@(LitInt n)       = e
renamer _ e@(LitDouble f)    = e
renamer _ e@(LitBool b)      = e
renamer _ e@(ImportData _)   = e
renamer f e@(Var n t)        = case lookupFM f n of
                                Nothing -> e
                                Just m  -> Var m t
renamer f (Lambda x ty e)    = Lambda x ty (renamer f e)  -- ??
renamer f (Let x e1 e2)      = Let x (renamer f e1) (renamer f e2)
renamer f (If e1 e2 e3)      = If (renamer f e1) (renamer f e2) (renamer f e3)
renamer _ e@Noop             = e
renamer f  (Loop x e1 e2)    = Loop x (renamer f e1) (renamer f e2)
renamer f (MkArr e1 x e2)    = MkArr (renamer f e1) x (renamer f e2)
renamer f (NewArr sz ty)     = NewArr (renamer f sz) ty
renamer f e                  = fromHnf (head, map (renamer f) args)
                              where (head,args) = toHnf e

testcse e = putStr (show (mkE e) ++ "\n\n" ++ show (mkE (cse e)))

cs1 = MkArr (LitInt 3) "k"
      (let x = (ReadArr (ReadArr (MkArr (LitInt 3) "i" (MkArr (LitInt 3) "j" (LitInt 3))) (Var "k" Int)) (Var "k" Int))
       in (Add (If (Var "a" Bool) x (LitInt 1)) 
                                   (If (Var "b" Bool) x (LitInt 2))))
cs12 = let size = LitInt 10 
           x = MkArr size "i" (LitInt 3)
           z = Var "j" Int
           y = ReadArr x z
           p = If ((Not (LessThan z (LitInt 0))) `And` (LessThan z size))
               (Mul (If (Var "a" Bool) y  (LitInt 1)) (If (Var "b" Bool) y (LitInt 2)))
               (LitInt 3)
           b = If (Var "a" Bool) p (LitInt 1) `Add` If (Var "b" Bool) p (LitInt 2)
       in MkArr (LitInt 11) "k" b

cs0 = MkArr (LitInt 3) "k"
       (let x = (ReadArr (ReadArr (MkArr (LitInt 3) "i" (MkArr (LitInt 3) "j" (LitInt 3))) (LitInt 1)) (LitInt 2))
       in (Add (If (Var "a" Bool) x (LitInt 1)) 
                                   (If (Var "b" Bool) x (LitInt 2))))

cs0a = MkArr (LitInt 3) "k"
       (let x = (ReadArr (ReadArr (MkArr (LitInt 3) "i" (MkArr (LitInt 3) "j" (LitInt 3))) (LitInt 1)) (LitInt 2))
       in x)


cs0b = MkArr (LitInt 2) "i" (MkArr (LitInt 3) "k" ((Add (Add (Var "k" Int) (Var "k" Int)) 
                                                        (Add (Var "i" Int) (Var "i" Int)))))
           
cs0c = MkArr (LitInt 2) "i" (MkArr (LitInt 3) "k" (If (Var "a" Bool)
                                                        (Add (Add (Var "k" Int) (Var "k" Int)) 
                                                        (Add (ReadArr (MkArr (LitInt 4) "j" (LitInt 5))
                                                                      (Var "i" Int)) 
                                                             (Var "i" Int))) (LitInt 0)))
           

           







