
module PPDExp where

import DExp
import Types
import BasicTypes
import PP

-- We could define the following Show instance, but let's not, so we can
-- see the structure when necessary, by means of the derived Show instance.

-- instance Show Exp where
--   showsPrec = showsPrecDExp

showsPrecDExp p e = (showPPDoc (PPDExp.ppExp e) p ++)

ppExp :: Exp -> PPDoc
ppExp e =
  case e of
    LitInt n     -> text (show n)
    LitDouble x  -> text (show x)
    LitBool b    -> text (show b)
    --LitF32vec4 v -> text (show v)
    ImportData s -> text ("importData " ++ show s)
    Var n ty     -> text n
    Lambda x ty e -> ppLambda x ty e
    If c a b     -> 
       ppOperator (0, NAssoc) $  -- because "else" is loose
         hang (text "if" <+> ppExp c <+> text "then")
           2  (ppExp a) $$
         sep [text "else", nest 2 (ppExp b)]
    Add e1 (LitInt n) | n < 0 -> ppOp (ppBinOp "-" e1 (LitInt (-n)))
    Add (LitInt n) e1 | n < 0 -> ppOp (ppBinOp "-" e1 (LitInt (-n)))
    Add e1 (LitDouble n) | n < 0 -> ppOp (ppBinOp "-" e1 (LitDouble (-n)))
    Add (LitDouble n) e1 | n < 0 -> ppOp (ppBinOp "-" e1 (LitDouble (-n)))
    Add e1 (Negate e2)     -> ppOp (ppBinOp "-" e1 e2)
    Add (Negate e2) e1     -> ppOp (ppBinOp "-" e1 e2)
    Add e1 e2              -> ppOp (ppBinOp "+" e1 e2)
    Mul e1 (Recip e2)      -> ppOp (ppBinOp "/" e1 e2)
    Mul e1 e2              -> ppOp (ppBinOp "*" e1 e2)
    Pow e1 (LitDouble 0.5) -> ppApply "sqrt" [e1]
    Pow e1 e2              -> ppOp (ppBinOp "**" e1 e2)
    LogBase e1 e2          -> ppApply "logBase" [e1,e2]
    Negate e               -> ppApply "negate" [e]
    Recip e                -> ppApply "recip" [e]
    Sin e                  -> ppApply "sin" [e]
    Cos e                  -> ppApply "cos" [e]
    Atan2 e1 e2            -> ppApply "atan2" [e1,e2]
    I2f e                  -> ppApply "i2f" [e]
    F2i e                  -> ppApply "f2i" [e]
    B2i e                  -> ppApply "b2i" [e]
    Abs e                  -> ppApply "abs" [e]
    Mod e1 e2              -> ppOp (ppBinOp "`mod`" e1 e2)
    Div e1 e2              -> ppOp (ppBinOp "`div`" e1 e2)
    Equal e1 e2            -> ppOp (ppBinOp "==" e1 e2)
    LessThan e1 e2         -> ppOp (ppBinOp "<" e1 e2)
    Min e1 e2              -> ppApply "min" [e1, e2]
    Max e1 e2              -> ppApply "max" [e1, e2]
    And e1 e2              -> ppOp (ppBinOp "&&" e1 e2)
    Or e1 e2               -> ppOp (ppBinOp "||" e1 e2)
    Not e	           -> ppApply "not" [e]
    LShift e1 (Negate e2)  -> ppOp (ppBinOp ">>>" e1 e2)
    LShift e1 (LitInt n) 
      | n < 0              -> ppOp (ppBinOp ">>>" e1 (LitInt (-n)))
    LShift e1 e2           -> ppOp (ppBinOp "<<<" e1 e2)
    RShift e1 e2           -> ppOp (ppBinOp ">>>" e1 e2)
    PackF32 e1 e2 e3 e4    -> ppApply "packF32" [e1,e2,e3,e4]
    NthF32 e1 e2           -> ppApply "nthF32" [e1,e2]
    SumF32 e               -> ppApply "sumF32" [e]
    MkTuple es             -> ppApply (mkTupleFun  (length es)) es
    TupleElt n e           -> ppApply (tupleEltFun (length tys) n) [e]
                                where Tuple tys = typeOf e
    BitAnd e1 e2           -> ppOp (ppBinOp ".&." e1 e2)
    BitOr e1 e2            -> ppOp (ppBinOp ".|." e1 e2)
    BitNot e               -> ppApply "complement" [e]
    Deref8 e               -> ppApply "deref8" [e]
    Deref32 e              -> ppApply "deref32" [e]
    Noop                   -> text "noop"
    Bind e1 (Lambda x ty body) -> 
       ppOperator (0, NAssoc) $
       ppExp e1 <+> text (">>= \\" ++ varsName ++ " ->") $$
       ppExp body
      where
	(vars, body') = stripLambdas [x] body

        varsName = concat [' ':var | var <- vars]
       
    Bind e1 e2             -> ppOp (ppBinOp ">>=" e1 e2)
    Return e		   -> ppApply "return" [e]
    Perform e              -> ppApply "performAction" [e]
    SetInt e1 e2           -> ppApply "setInt" [e1, e2]
    Loop v m body          -> ppLoop v m body
    ReadArr arr ix         -> ppApply "readArr" [arr,ix]
    ReadMutArr arr ix      -> ppApply "readMutArr" [arr,ix]
    WriteArr arr ix el     -> ppApply "writeArr" [arr,ix,el]
    NewArr size ty         -> ppApply "newArr" [size]
    MkArr size i body      -> ppMkArr size i body
    Rand                   -> text "rand"
    NewRef init            -> ppApply "newRef" [init]
    ReadRef ref            -> ppApply "readRef" [ref]
    WriteRef ref newVal    -> ppApply "writeRef" [ref,newVal]
    DoProgress val max     -> ppApply "doProgress" [val, max]
    Copy1DArray base len   -> ppApply "copy1DArray" [base,len]
    e@(Let i e1 e2)        -> ppOperator (0,NAssoc) (text "let" $$ ppLet e)
     where
      ppOne i e                       = nest 1 (text i <+> char '=' <+> ppExp e)
      ppLet (Let i e1 e2@(Let _ _ _)) = ppOne i e1 $$ ppLet e2
      ppLet (Let i e1 e2)             = ppOne i e1 $$ text "in " $$ ppExp e2
    _                      -> error ("PPDExp.ppExp: no case for " ++ show e)
 where
  ppOp = ppOperator (opInfo e)

mkTupleFun :: Int -> String
mkTupleFun 2 = "pair"
mkTupleFun n = "tuple" ++ show n

tupleEltFun :: Int -> Int -> String
tupleEltFun 2    0 = "fst"
tupleEltFun 2    1 = "snd"
tupleEltFun leng i = mkTupleFun leng ++ "_" ++ show i

ppBinOp :: String -> Exp -> Exp -> PPDoc
ppBinOp op e1 e2 = sep [ppExp e1, text op, onRight (ppExp e2)]

ppLambda var _ body =
  ppOperator (0, NAssoc) $
    text ("\\" ++ varsName ++ " ->") $$ nest 3 (ppExp body')
 where
   (vars, body') = stripLambdas [var] body

   varsName = concat [' ':var | var <- vars]


stripLambdas vars (Lambda var' _ body') =
  stripLambdas (var':vars) body'

stripLambdas vars e = (reverse vars, e)


ppLoop :: Id -> Exp -> Exp -> PPDoc
ppLoop var max body =
  ppOperator (apPrec, NAssoc) $ 
     hang (text "loop" <+> ppExp max <+> text ("(\\ " ++ var ++ " -> "))
        2 (ppOperator (0, NAssoc) (ppExp body)) <+> text ")"

ppMkArr size i body =
  ppOperator (apPrec, NAssoc) $ 
     hang (text "mkArr" <+> ppExp size <+> text ("(\\ " ++ i ++ " -> "))
        2 (ppOperator (0, NAssoc) (ppExp body)) <+> text ")"

ppApply :: String -> [Exp] -> PPDoc
ppApply f args = ppOperator (apPrec,LAssoc) (text f <+> onRight (sep (map ppExp args)))
