%
%
%

Representing Colors as a vector of 4 floats (which maps directly
onto PIII's "Streaming SIMD extensions" 128-bit xmm registers.)

\begin{code}
module ColorSSE
        (
	 ColorE

	-- aToColor (a,r,g,b) = (a,a,a,a)
	, aToColor     -- :: ColorE -> ColorE

	, bgra         -- :: FloatE -> FloatE -> FloatE -> FloatE -> ColorE
        , colorToBGRA  -- :: ColorE -> (FloatE,FloatE,FloatE,FloatE)
	
	, fractToColor -- :: FractionE -> ColorE

        , BGR24 , BGR24E
        , RGB24 , RGB24E                -- photoshop

        -- Int <-> Color conversions:
	, fromBGR32    -- :: BGR32E -> ColorE
	, fromBGR24    -- :: BGR24E -> ColorE
	, fromRGB24    -- :: IntE -> ColorE

	, toBGR24      -- :: ColorE -> BGR24E
	, toRGB24      -- :: ColorE -> BGR24E

        , distC, distSqC  -- :: ColorE -> ColorE -> FloatE

	  -- basic color combinators:
	, negC	       -- :: ColorE -> ColorE
	, lerpC        -- :: FractionE -> ColorE -> ColorE -> ColorE
	, fadeC	       -- :: FractionE -> ColorE -> ColorE
	, overC        -- :: ColorE -> ColorE -> FloatE

	, usesSSE 
        , colorVar     -- :: String -> ColorE
	) where

import Exp
import Bits
import Int
\end{code}

To exploit PentiumIII extensions, represent a colour as a
four tuple of floats (F32vec4):

Memory order is B,G,R,A, as an extension of BMP format.  In keeping the
the little-endianness of Intel, however, the order within a register is
A,R,G,B, i.e., the high order float is A and low order is B.  The
nthF32 function follows memory order.

The packF32 function, which mirrors the C++ F32vec4 constructor,
reverses its arguments, relative to the nthF32 function, which mirrors
the C++ v[i] operator.  That is, nth i (packF32 x3 x2 x1 x0) == xi.
For sanity, we want the subscripting order to mirror the order in a bmp
file, i.e., bgr.  Thus argb and packF32 want the same order.

Sorry to be so verbose, but this is the sort of thing really confuses me!

\begin{code}
type ColorE = F32vec4E


-- To signal to the backend whether we're using SSE (=> use the Intel C
-- compiler), all implementations of Color export the following flag:
usesSSE :: Bool
usesSSE = True
\end{code}


\begin{code}
bgra :: FloatE -> FloatE -> FloatE -> FloatE -> ColorE
bgra b g r a = packF32 a r g b

alpha :: ColorE -> FractionE
alpha = colorA

fractToColor :: FractionE -> ColorE
fractToColor = replF32

aToColor :: ColorE -> ColorE
aToColor c = replF32 (colorA c)

colorA :: ColorE -> FractionE
colorA = nthF32 3

colorToBGRA :: ColorE -> (FloatE,FloatE,FloatE,FloatE)
colorToBGRA c = (nthF32 0 c, nthF32 1 c, nthF32 2 c, nthF32 3 c)

\end{code}

Some useful F32vec4E functions and values:
\begin{code}
-- local helper; not exported (but the v. similar looking fractToColor is.)
replF32 :: FloatE -> F32vec4E
replF32 x = packF32 x x x x

-- Some useful constants.  They're defined here, globally to help
-- pointer-sharing, upon which CSE is based.
vec255,vecRecip255, vecHalf :: F32vec4E
vec255      = replF32 255
vecRecip255 = recip vec255
vecHalf     = replF32 0.5
\end{code}


Color conversions:

\begin{code}

fromBGR32 :: BGR32E -> ColorE
fromBGR24 :: BGR24E -> ColorE
fromRGB24 :: RGB24E -> ColorE

fromBGR32 nCol = vecRecip255 * byteQuad
 where
   -- Note the pack/memory order reversal.
   byteQuad = packF32 255 (extract 2) (extract 1) (extract 0)
   -- Extract nth byte, with low = byte 0.  Convert to float
   extract n = i2f ((nCol >>> (8*n)) .&. 0xff)

fromBGR24 = fromBGR32

fromRGB24 nCol = vecRecip255 * byteQuad
 where
   -- Note the pack/memory order reversal.
   byteQuad = packF32 255 (extract 0) (extract 1) (extract 2)
   -- Extract nth byte, with low = byte 0.  Convert to float
   extract n = i2f ((nCol >>> (8*n)) .&. 0xff)

toBGR24 :: ColorE -> BGR24E
toRGB24 :: ColorE -> RGB24E

toBGR24 = toBytePacked 3 False
toRGB24 = toBytePacked 3 True

toBytePacked :: Int -> Bool -> ColorE -> IntE
toBytePacked n revBytes col =
  -- Scale and prepare for rounding.
  let scaledCol = vec255 * col + vecHalf in
    -- Extract each part, convert to int, shift into place, and add.
    sum [ f2i (nthF32 j' scaledCol) <<< (8*i')
        | i <- [0 .. n-1]
        , let i' = fromInt i
        , let j' | revBytes  = fromInt (n - 1 - i)
                 | otherwise = i'
        ]

\end{code}

Color combinators that are 'better' (aka. more efficiently) expressed
in terms of the concrete Color representation.

\begin{code}
-- Color negative.
negC :: ColorE -> ColorE
negC col = bgra a a a (2*a) - col
 where a = colorA col

  -- Note: we're cheating a little here, since the bgra application
  -- above may pass an alpha component that is > 1.

-- Dot product for colors
dotC :: ColorE -> ColorE -> FloatE
dotC c c' = sumF32 (c*c')

-- Distance and distance squared in BGRA space
distC, distSqC :: ColorE -> ColorE -> FloatE

distSqC c c' = dotC d d where d = c-c'

distC c c' = sqrt (distSqC c c')


lerpC :: FractionE -> ColorE -> ColorE -> ColorE
lerpC x a b = a + fractToColor x * (b - a)

-- Fade by a given factor x, which should be between 0 and 1.  fade 1 is
-- the identity, and fade 0 makes invisible.  Hmm.. possibly misleading
-- name.
fadeC x col = fractToColor x * col

-- Overlay function on two colors
overC :: ColorE -> ColorE -> ColorE
overC top bot =
  top + (fractToColor 1 - aToColor top) * bot
  -- Equivalent, but more efficient.
  -- top + replF32 (1 - colorA top) * bot

-- For testing

-- Colors are expressions
colorVar :: String -> ColorE
colorVar = var

\end{code}
